#  Copyright (C) 2000 FlicNET and Mike Ivanov <mike@flic.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

import time, random

import ZODB
import Globals
import Persistence

from Products.ZPatterns.Specialists import Specialist
from Products.ZPatterns.Rack import Rack, RackMountable
from Products.ZPatterns.PlugIns import PlugInGroup
from Products.ZPatterns.DataManagers import DataManager


################################################################
#
# Session
#

class Session( RackMountable) :
    """
    Generic Session Class.
    Usage in particular tasks:
       1. Create Z Class based on the 'Z Session'
       2. Define methods you need
    """
    
    meta_type = "Z Session"

    isAZSession = 1

    def idleTime( self) :
        """The time since last touch call"""
        return time.time() - self.lastAccessed

    def touch( self) :
        """Sets touch mark to current time"""
        self.lastAccessed = time.time()


class LazySession( Session) :
    """
    Use get/set calls to handle attributes
    """
    
    meta_type = "Lazy Z Session"

    isALazyZSession = 1

    # next methods work on 'Basic' propertysheet

    def get( self, id, default=None) :
        """
        Returns value of property 'id' or 'default' if there is
        no such property.
        """
        return self.basicSheet().getProperty( id, default)

    def has( self, id) :
        """1 if property 'id' exists"""
        return self.basicSheet().hasProperty( id)

    def add( self, id, value, type='string') :
        """Adds new property 'id' of type 'type' with initial value 'value'"""
        self.basicSheet().manage_addProperty( id, value, type)

    def update( self, **kw) :
        """Updates whole set of properties"""
        apply( self.basicSheet().manage_changeProperties, (), kw)

    def set( self, id, value) :
        """Sets particular property"""
        apply( self.update, (), {id:value})
            
    def remove( self, ids) :
        """Removes property 'id'"""
        self.basicSheet().manage_delProperties( ids)

    def ids( self) :
        """Returns list of property ids"""
        return self.basicSheet().propertyIds()

    def basicSheet( self) :
        """Assumes propertysheet 'Basic' exists and returns it"""
        sheet = self.propertysheets.get('Basic')
        if sheet is None :
            self.propertysheets.manage_addPropertySheet('Basic','Basic')
            sheet = self.propertysheets.get('Basic')
        return sheet


################################################################
#
# SessionManager
#

class SessionTimer( Persistence.Persistent) :
    """One-field space saver"""

    def __init__( self) :
        """ """
        self.time = time.time()

    def idleTime( self) :
        """ """
        return time.time() - self.time

    def touch( self) :
        """ """
        self.time = time.time()


class Counter( Persistence.Persistent) :
    """One-field space saver"""

    def __init__( self) :
        """ """
        self._last = -1

    def next( self) :
        """ """
        n = self._last = self._last + 1
        return n


class SessionNotFoundException( LookupError) : pass


class SessionManager( Specialist) :
    """ """
    meta_type = "Z Session Manager"

    manage_options_left = Specialist.manage_options_left + (
	{'label':'Contents', 'action':'manage_sessionsForm'},
	)

    _properties = (
        {'id':'title',       'type':'string',  'mode':'w'},
        {'id':'session_ttl', 'type':'int',     'mode':'w'},
        {'id':'cookie_name', 'type':'string',  'mode':'w'},
        {'id':'cookie_path', 'type':'string',  'mode':'w'},
        {'id':'auto_create', 'type':'boolean', 'mode':'w'},
        {'id':'auto_remove', 'type':'boolean', 'mode':'w'},
        )

    __ac_permissions__ = (
        ('View', (
            'getItem', 'getSession', 'newItem', 'newSession',
            'currentSession', '__call__',
            )),
	('Change permissions', (
            'manage_access',
            )),
	('Manage Z Sessions', (
           'removeDead', 'getSessionIds',
           'manage_sessionsForm', 'manage_deleteSessions', 'manage_removeDead',
           )),
	)

    SessionSourcesGroup = PlugInGroup(
        'SessionSourcesGroup', ['Z Session Source'],
        attr='rackList', title='Session Sources'
        )

    __plugin_groups__ = (SessionSourcesGroup,) + DataManager.__plugin_groups__

    #---------------------------------------------------------

    def __init__(
        self, id, title = '', session_ttl = 60*60*12,
        cookie_name = 'ZSESSION%08X%04X%04X' % (
            time.time(), random.randint(0,0xffff), random.randint( 0, 0xffff)
            ),
        cookie_path = '/',
        auto_create = 1, auto_remove = 1
    ) :
	""" """
        self._timer = SessionTimer()

        self.id = id
	self.title = title
        self.session_ttl = session_ttl
        self.cookie_name = cookie_name
        self.cookie_path = cookie_path
        self.auto_create = auto_create
        self.auto_remove = auto_remove


    # external API (responsibilities) ---------------------------------

    def currentSession( self, REQUEST) :
        """Returns 'current' session for REQUEST's brouser"""
        if self.auto_remove :
            if self._timer.idleTime() > self.session_ttl / 2 :
                self.removeDead()
                
        id = REQUEST.get( self.cookie_name, None)
        if id is not None :
            id = self._decrypt( id)
            session = self.getSession( id)
            if session is None :
                if self.auto_create :
                    session = self.newSession()
                else :
                    raise SessionNotFoundException
            else :
                session.touch()
        else :
            session = self.newSession()

        if id != session.id :
            id = self._encrypt( session.id)
            REQUEST.set( self.cookie_name, id)
            REQUEST.response.setCookie( self.cookie_name, id, path=self.cookie_path)

        return session

    def __call__( self) :
        """ """
        return self.currentSession( self.REQUEST)

    getSession = Specialist.getItem

    def newItem( self) :
        """ """
        session = Specialist.newItem( self)
        session.created = session.lastAccessed = time.time()
        session.host = self.REQUEST.get('REMOTE_ADDR', 'UNKNOWN')
        return session

    newSession = newItem

    # kinda protected ---------------------------------------

    def removeDead( self) :
        """Removes sessions whose idleTime > TTL"""

        for id in self._getSource().findDeadSessions() :
            sess.manage_delete()
        self._timer.touch()

    def findDeadSessions( self) :
        """
        Put method with the same name in SessionSource if you think
        you can do this better :-)
        """
        dead = []
        for id in self._getSource().getSessionIds() :
            sess = self.getItem( id)
            if sess.idleTime() > self.session_ttl :
                dead.append( sess)
        return dead

    def getSessionIds( self) :
        """Returns list of persistent session ids"""
        return list( self._getSource().getPersistentItemIDs())

    # internals --------------------------------------------------

    def newKey( self) :
        """Default key generator"""

        if hasattr( self, 'generateKey') :
            return self.generateKey( self, self.REQUEST)
        if not hasattr( self, '_counter') :
            self._counter = Counter()
        return "%012d" % self._counter.next()

    def _getSource( self) :
        """ """
        return self.rackList[0].__of__(self)

    def _encrypt( self, data) :
        """ """
        if hasattr( self, 'encryptString') :
            return self.encryptString( self, self.REQUEST, data=data)
        return data

    def _decrypt( self, data) :
        """ """
        if hasattr( self, 'decryptString') :
            return self.decryptString( self, self.REQUEST, data=data)
        return data


    
manage_addSessionManagerForm = Globals.HTMLFile(
    'addSessionManager', globals(),
    SessionSourcesMetaTypes = SessionManager.SessionSourcesGroup.all_meta_types,
    )

def manage_addSessionManager(
    self, id, title, session_ttl, cookie_name, cookie_path,
    source, auto_create=0, auto_remove=0, REQUEST=None
    ) :
    
    """Ummgh..."""

    ob = SessionManager(
        id, title=title, session_ttl=session_ttl, cookie_name=cookie_name,
        cookie_path=cookie_path, auto_create=auto_create,
        auto_remove=auto_remove
        )
    
    id = self._setObject( ob.id, ob)
    ob = getattr( self, id)

    ob.SessionSourcesGroup._constructPlugIn(
        source, id='SessionSource', title='Default Z Session Source'
        )

    ob.manage_addDTMLMethod(
        id='index_html', file=html_index_html
        )

    ob.manage_addDTMLMethod(
        id='manage_sessionsForm', file=html_manage_sessionsForm
        )

    ob.manage_addDTMLMethod(
        id='manage_deleteSessions', file=html_manage_deleteSessions
        )

    ob.manage_addDTMLMethod(
        id='manage_removeDead', file=html_manage_removeDead
        )

    # Success!
    if REQUEST: return self.manage_main( self, REQUEST, update_menu=1)


html_index_html = '''\
<html>
<dtml-with "Session()">
  session id = <dtml-var id><br>
  <dtml-try>
  has xxx? <dtml-var "has('xxx')"><br>
  <dtml-if "not has('xxx')">
    adding xxx...
    <dtml-call "add('xxx',['XXX'],'list')">
    added<br>
  <dtml-else>
    now xxx = <dtml-var "get('xxx')"><br>
    <dtml-if "get('xxx') == 'XXX'">
      <dtml-call "update( xxx='YYY')">
    <dtml-else>
      <dtml-call "update( xxx='XXX')">
    </dtml-if>
  </dtml-if>
  <dtml-except>
    Set up Lazy Z Session as default ZClass in SessionSource please
  </dtml-try>
</dtml-with>
'''

html_manage_sessionsForm = '''\
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.0 Transitional//EN" "http://www.w3.org/TR/REC-html40/loose.dtd">
<HTML lang="en">
<HEAD>
    <TITLE>Z Session Manager Contents</TITLE>
</HEAD>
<BODY BGCOLOR="#FFFFFF" LINK="#000099" VLINK="#555555">

<dtml-var manage_tabs>

<H2>Sessions</H2>

<dtml-call "REQUEST.set('sessions', getSessionIds())">
<dtml-in sessions>
        <dtml-if sequence-start>
            <form action="." method=post>
            <p>Current session TTL is <dtml-var session_ttl> sec.</p>
            <table border=1>
                <tr>
                    <th></th><th>Id</th><th>Host</th>
                    <th>Created</th><th>Idle (sec)</th>
                </tr>
        </dtml-if>
        <dtml-with "getSession(_['sequence-item'])">
        <tr>
            <td><input type=checkbox name=ids:list value="&dtml-id;"></td>
            <td>&dtml-id;</td>
            <td>&dtml-host;</td>
            <td><dtml-var "ZopeTime( created)"></td>
            <dtml-call "REQUEST.set('dead', idleTime() > session_ttl)">
            <td>
                <dtml-if dead><font color=red></dtml-if>
                &dtml-idleTime;
                <dtml-if dead></font></dtml-if>
            </td>
        </tr>
        </dtml-with>
        <dtml-if sequence-end>
            </table>
            <input type=submit name=manage_deleteSessions:method
                value=" Delete Selected "
            >
            <input type=submit name=manage_removeDead:method
                value=" Remove Dead Sessions "
            >
            </form>
        </dtml-if>
<dtml-else>
    No sessions
</dtml-in>

</BODY>
</HTML>
'''

html_manage_deleteSessions = '''\
<html>
<dtml-in ids>
   <dtml-with "getSession(_['sequence-item'])">
       <dtml-call manage_delete>
   </dtml-with>
</dtml-in>
<dtml-call "RESPONSE.redirect('manage_sessionsForm')">
Done
'''

html_manage_removeDead = '''\
<html>
<dtml-call removeDead>
<dtml-call "RESPONSE.redirect('manage_sessionsForm')">
Done
'''


####################################################################
#
# SessionSource
#

# I just did not know how to restrict standard Rack

class SessionSource( Rack) :
    """ """
    meta_type = "Z Session Source"

    _defaultClass = Session
    __plugin_kind__ = 'Z Session Source'

    def _zclassOK( self, z) :
	""" """
	return getattr( z._zclass_, 'isAZSession', 0)


manage_addSessionSourceForm = Globals.HTMLFile(
    'addSessionSource', globals()
    )

def manage_addSessionSource( self, id, title='', REQUEST=None) :
    """ """
    ob = SessionSource( id, title)
    return self.Destination()._installPlugIn(
	ob, 'Added Session Source.', REQUEST
	)









